from enum import IntEnum


class ShaderCompilerPlatform(IntEnum):
    kShaderCompPlatformNone = -1
    kShaderCompPlatformGL = 0
    kShaderCompPlatformD3D9 = 1
    kShaderCompPlatformXbox360 = 2
    kShaderCompPlatformPS3 = 3
    kShaderCompPlatformD3D11 = 4
    kShaderCompPlatformGLES20 = 5
    kShaderCompPlatformNaCl = 6
    kShaderCompPlatformFlash = 7
    kShaderCompPlatformD3D11_9x = 8
    kShaderCompPlatformGLES3Plus = 9
    kShaderCompPlatformPSP2 = 10
    kShaderCompPlatformPS4 = 11
    kShaderCompPlatformXboxOne = 12
    kShaderCompPlatformPSM = 13
    kShaderCompPlatformMetal = 14
    kShaderCompPlatformOpenGLCore = 15
    kShaderCompPlatformN3DS = 16
    kShaderCompPlatformWiiU = 17
    kShaderCompPlatformVulkan = 18
    kShaderCompPlatformSwitch = 19
    kShaderCompPlatformXboxOneD3D12 = 20
